// PrivIoXP.c - create devices for parallel port direct I/O access

#include "ntddk.h"
#include "ntiologc.h"
#include "parallel.h"
#include "PrivIoXP.h"
#include "../PrivIoXP_IOCTL.h"

// local I/O permissions bitmap
#define IOPM_SIZE 0x2000
typedef UCHAR IOPM[IOPM_SIZE];
IOPM *IOPM_local = NULL;

// undocumented Windows kernel routines
void Ke386IoSetAccessProcess(PEPROCESS, int);
void Ke386QueryIoAccessMap(int, IOPM *);
void Ke386SetIoAccessMap(int, IOPM *);

// forward references
NTSTATUS PrivIoXPCleanup(IN PDEVICE_OBJECT DeviceObject, IN PIRP pIrp);
NTSTATUS PrivIoXPCreate(IN PDEVICE_OBJECT DeviceObject, IN PIRP pIrp);
NTSTATUS PrivIoXPRead(IN PDEVICE_OBJECT DeviceObject, IN PIRP pIrp);
NTSTATUS PrivIoXPClose(IN PDEVICE_OBJECT DeviceObject, IN PIRP pIrp);
NTSTATUS PsLookupProcessByProcessId(IN ULONG ulProcId,
                                    OUT struct _EPROCESS **pEProcess);
NTSTATUS PrivIoXPIOCtl(IN PDEVICE_OBJECT DeviceObject, IN PIRP pIrp);
VOID PrivIoXPUnload(IN PDRIVER_OBJECT DriverObject);

//
//  Driver CreateDispatch handler
//
NTSTATUS PrivIoXPCreateDispatch(IN PDEVICE_OBJECT DeviceObject, IN PIRP pIrp)
{
    pIrp->IoStatus.Information = 0;
    pIrp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest(pIrp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;
}

//
//  Log stuff into "event log" for debug purposes
//
VOID PrivIoXPLogInfo(
    IN PDRIVER_OBJECT DrvObj,  // can be NULL
    IN PDEVICE_OBJECT DevObj,  // can be NULL
    IN NTSTATUS Code,          // driver-defined code (see "ntiologc.h")
    IN NTSTATUS FinalStatus)   // driver exit code
{
    PIO_ERROR_LOG_PACKET LogEntry = NULL;

    if (DrvObj != NULL) {
        LogEntry = IoAllocateErrorLogEntry(DrvObj, sizeof(IO_ERROR_LOG_PACKET));
    } else if (DevObj != NULL){
        LogEntry = IoAllocateErrorLogEntry(DevObj, sizeof(IO_ERROR_LOG_PACKET));
    }
    if (!LogEntry) return;

    LogEntry->RetryCount = 0;
    LogEntry->DumpDataSize = 0;
    LogEntry->NumberOfStrings = 0;
    LogEntry->StringOffset = 0;
    LogEntry->ErrorCode = Code;
    LogEntry->UniqueErrorValue = 0;
    LogEntry->FinalStatus = FinalStatus;
    LogEntry->SequenceNumber = 0;
    IoWriteErrorLogEntry(LogEntry);
    return;
}

//
//  Make port/class/link Unicode name strings for new device
//
BOOLEAN PrivIoXPMakeNames(
    IN ULONG ulParPortNum,             // passed port number (1..n)
    OUT PUNICODE_STRING pusPortName,   // returned port name
    OUT PUNICODE_STRING pusClassName,  // returned class name
    OUT PUNICODE_STRING pusLinkName    // returned linkage name
){
    UNICODE_STRING usLinkPrefix, usPortSuffix, usClassSuffix, usPrefix;
    UNICODE_STRING usDigits, usLinkDigits, usLinkSuffix;
    WCHAR wcDigitsBuffer[10], wcLinkDigitsBuffer[10], wcLinkSuffixBuffer[64];
    ANSI_STRING asLinkSuffixAnsi;
    NTSTATUS Status;

    RtlInitUnicodeString(&usPortSuffix, DD_PARALLEL_PORT_BASE_NAME_U);
    RtlInitUnicodeString(&usClassSuffix, L"ParallelIO");
    RtlInitUnicodeString(&usPrefix, L"\\Device\\");
    RtlInitUnicodeString(&usLinkPrefix, L"\\DosDevices\\");

    // convert simple string to Unicode string ("PPORTIO" -> L"PPORTIO")
    RtlInitAnsiString(&asLinkSuffixAnsi, PRIVIOXP_DEVICE_BASE_NAME);
    usLinkSuffix.Length = 0;
    usLinkSuffix.MaximumLength = sizeof(wcLinkSuffixBuffer) / sizeof(WCHAR);
    usLinkSuffix.Buffer = wcLinkSuffixBuffer;
    Status = RtlAnsiStringToUnicodeString(
                                  &usLinkSuffix, &asLinkSuffixAnsi, FALSE);
    if (!NT_SUCCESS(Status)) return (FALSE);

    // convert ulParPortNum to Unicode string
    usDigits.Buffer = wcDigitsBuffer;
    usDigits.Length = 0;
    usDigits.MaximumLength = sizeof(wcDigitsBuffer);
    Status = RtlIntegerToUnicodeString(ulParPortNum, 10, &usDigits);
    if (!NT_SUCCESS(Status)) return (FALSE);

    // convert ulParPortNum + 1 to Unicode string
    usLinkDigits.Buffer = wcLinkDigitsBuffer;
    usLinkDigits.Length = 0;
    usLinkDigits.MaximumLength = sizeof(wcLinkDigitsBuffer);
    Status = RtlIntegerToUnicodeString(ulParPortNum + 1, 10, &usLinkDigits);
    if (!NT_SUCCESS(Status)) return (FALSE);

    // initialize/allocate storage for port name string
    pusPortName->MaximumLength = usPrefix.Length + usPortSuffix.Length +
                                           usDigits.Length + sizeof(WCHAR);
    pusPortName->Length = 0;
    pusPortName->Buffer = ExAllocatePool(PagedPool, pusPortName->MaximumLength);
    if (!pusPortName->Buffer) return (FALSE);

    // make port name string
    RtlZeroMemory(pusPortName->Buffer, pusPortName->MaximumLength);
    RtlAppendUnicodeStringToString(pusPortName, &usPrefix);
    RtlAppendUnicodeStringToString(pusPortName, &usPortSuffix);
    RtlAppendUnicodeStringToString(pusPortName, &usDigits);

    // initialize/allocate storage for class name string
    pusClassName->Length = 0;
    pusClassName->MaximumLength = usPrefix.Length + usClassSuffix.Length +
                                              usDigits.Length + sizeof(WCHAR);
    pusClassName->Buffer = ExAllocatePool(
                                      PagedPool, pusClassName->MaximumLength);
    if (!pusClassName->Buffer) {
        ExFreePool(pusPortName->Buffer);
        return (FALSE);
    }

    // make class name string
    RtlZeroMemory(pusClassName->Buffer, pusClassName->MaximumLength);
    RtlAppendUnicodeStringToString(pusClassName, &usPrefix);
    RtlAppendUnicodeStringToString(pusClassName, &usClassSuffix);
    RtlAppendUnicodeStringToString(pusClassName, &usDigits);

    // initialize/allocate storage for link name string
    pusLinkName->Length = 0;
    pusLinkName->MaximumLength = usLinkPrefix.Length + usLinkSuffix.Length +
        usLinkDigits.Length + sizeof(WCHAR);
    pusLinkName->Buffer = ExAllocatePool(PagedPool, pusLinkName->MaximumLength);
    if (!pusLinkName->Buffer) {
        ExFreePool(pusPortName->Buffer);
        ExFreePool(pusClassName->Buffer);
        return (FALSE);
    }

    // make link name string
    RtlZeroMemory(pusLinkName->Buffer, pusLinkName->MaximumLength);
    RtlAppendUnicodeStringToString(pusLinkName, &usLinkPrefix);
    RtlAppendUnicodeStringToString(pusLinkName, &usLinkSuffix);
    RtlAppendUnicodeStringToString(pusLinkName, &usLinkDigits);

    return (TRUE);
}

//
//  Get port info from port device and store in device extension
//
NTSTATUS PrivIoXPGetInfoFromPortDevice(IN OUT PDEVICE_EXTENSION extension)
{
    PARALLEL_PORT_INFORMATION portInfo;
    PIRP pIrp;
    KEVENT kEvent;
    IO_STATUS_BLOCK ioStatus;
    NTSTATUS Status;

    // initialize event for IOCTL
    KeInitializeEvent(&kEvent, NotificationEvent, FALSE);

    // build IRP to get port info and send to driver
    pIrp = IoBuildDeviceIoControlRequest
                     (IOCTL_INTERNAL_GET_PARALLEL_PORT_INFO,
                     extension->PortDeviceObject, NULL, 0, &portInfo,
                     sizeof(PARALLEL_PORT_INFORMATION), TRUE, 
                     &kEvent, &ioStatus);
    if (!pIrp) {
        return (STATUS_INSUFFICIENT_RESOURCES);
    }
    Status = IoCallDriver(extension->PortDeviceObject, pIrp);
    if (!NT_SUCCESS(Status)) return (Status);

    // wait for request to complete
    Status = KeWaitForSingleObject(&kEvent, Executive, KernelMode, FALSE, NULL);
    if (!NT_SUCCESS(Status)) return (Status);

    // save port info in device extension
    extension->OriginalController = portInfo.OriginalController;
    extension->Controller = portInfo.Controller;
    extension->SpanOfController = portInfo.SpanOfController;
    extension->FreePort = portInfo.FreePort; // use to release port
    extension->FreePortContext = portInfo.Context;  // context to release

    // check of port span is large enough?
    if (extension->SpanOfController < PPORT_REGISTER_SPAN) {
        return (STATUS_INSUFFICIENT_RESOURCES);
    }
    return (Status);
}

//
//  Create new device object for one parallel port
//
VOID PrivIoXPCreateDevice(
    IN PDRIVER_OBJECT DriverObject,  // our driver object
    IN ULONG ulParPortNum            // parallel port number (1..n)
){
    UNICODE_STRING usPortName, usClassName, usLinkName;
    PDEVICE_OBJECT deviceObject;
    PDEVICE_EXTENSION extension;
    PFILE_OBJECT fileObject;
    NTSTATUS Status;

    // create strings for "port" and "class" names
    if (!PrivIoXPMakeNames(ulParPortNum,
                         &usPortName, &usClassName, &usLinkName)) {
        return;
    }

    // create new device
    Status = IoCreateDevice(DriverObject, sizeof(DEVICE_EXTENSION),
                            &usClassName, FILE_DEVICE_PARALLEL_PORT, 0,
                            FALSE, &deviceObject);
    if (!NT_SUCCESS(Status)) {
        ExFreePool(usLinkName.Buffer);
        goto finis;
    }

    // set and initialize device extension
    extension = deviceObject->DeviceExtension;
    RtlZeroMemory(extension, sizeof(DEVICE_EXTENSION));
    extension->DeviceObject = deviceObject;
    deviceObject->Flags |= DO_BUFFERED_IO;

    // get needed parallel port info for later IoCallDriver calls
    Status = IoGetDeviceObjectPointer(&usPortName, FILE_READ_ATTRIBUTES,
                                      &fileObject,
                                      &extension->PortDeviceObject);
    if (!NT_SUCCESS(Status)) {
        IoDeleteDevice(deviceObject);
        ExFreePool(usLinkName.Buffer);
        goto finis;
    }

    // release file object (we don't need it)
    ObDereferenceObject(fileObject);

    // set stack size (used by IRP)
    extension->DeviceObject->StackSize =
        extension->PortDeviceObject->StackSize + 1;

    // initialize our work queue
    InitializeListHead(&extension->WorkQueue);
    extension->CurrentIrp = NULL;

    // get port information (saved in device extension)
    Status = PrivIoXPGetInfoFromPortDevice(extension);
    if (!NT_SUCCESS(Status)) {
        IoDeleteDevice(deviceObject);
        ExFreePool(usLinkName.Buffer);
        goto finis;
    }

    // setup unprotected symbolic links for user-mode access
    Status = IoCreateUnprotectedSymbolicLink(&usLinkName, &usClassName);
    if (!NT_SUCCESS(Status)) {
        extension->CreatedSymbolicLink = FALSE;
        ExFreePool(usLinkName.Buffer);
        goto finis;
    }
    extension->CreatedSymbolicLink = TRUE;  // needed for cleanup
    extension->SymbolicLinkName = usLinkName;

  finis:
    ExFreePool(usPortName.Buffer);
    ExFreePool(usClassName.Buffer);
}

//
//  Driver DriverEntry handler
//
NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject,
                     IN PUNICODE_STRING RegistryPath)
{
    ULONG ulCount;

    // setup default I/O permissions
    IOPM_local = MmAllocateNonCachedMemory(sizeof(IOPM));
    if (IOPM_local == 0) {
        PrivIoXPLogInfo(DriverObject, NULL, IO_ERR_INSUFFICIENT_RESOURCES,
                        STATUS_INSUFFICIENT_RESOURCES);
        return STATUS_INSUFFICIENT_RESOURCES;
    }
    RtlFillMemory(IOPM_local, sizeof(IOPM), 0xff);  // no access

    // search for a parallel port and setup a new device
    for (ulCount = 0;
             ulCount < IoGetConfigurationInformation()->ParallelCount;
                                                               ulCount++) {
         PrivIoXPCreateDevice(DriverObject, ulCount);
    }

    // did we find a parallel port?
    if (!DriverObject->DeviceObject) {
        // "no device found" error!!
        PrivIoXPLogInfo(DriverObject, NULL, IO_ERR_CONFIGURATION_ERROR,
                        STATUS_NO_SUCH_DEVICE);
        return (STATUS_NO_SUCH_DEVICE);
    }

    // setup driver entry points for new device
    DriverObject->MajorFunction[IRP_MJ_CREATE] = PrivIoXPCreate;
    DriverObject->MajorFunction[IRP_MJ_CLOSE] = PrivIoXPClose;
    DriverObject->MajorFunction[IRP_MJ_READ] = PrivIoXPRead;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL] = PrivIoXPIOCtl;
    DriverObject->MajorFunction[IRP_MJ_CLEANUP] = PrivIoXPCleanup;
    DriverObject->DriverUnload = PrivIoXPUnload;

    // log "success" message to system "event log"
    PrivIoXPLogInfo(DriverObject, NULL, ((NTSTATUS)0x40070006L),
                        STATUS_SUCCESS);
    return (STATUS_SUCCESS);
}

//
//  Driver open (i.e., CreateFile) handler
//
NTSTATUS PrivIoXPCreate(IN PDEVICE_OBJECT DeviceObject, IN PIRP pIrp)
{
    PDEVICE_EXTENSION extension = DeviceObject->DeviceExtension;
    PIRP irp;
    KEVENT event;
    ULONG ioBase;
    ULONG ioSpan;
    ULONG ioAddr;
    ULONG offset;
    UCHAR mask;
    LARGE_INTEGER liTimeout;
    IO_STATUS_BLOCK ioStatus;
    NTSTATUS Status;

    // give direct I/O rights to parallel port
    Ke386IoSetAccessProcess(PsGetCurrentProcess(), 1);
    Ke386QueryIoAccessMap(1, IOPM_local);
    RtlFillMemory(IOPM_local, sizeof(IOPM), 0xff);  // no access
    ioBase = (ULONG)(extension->Controller);
    ioSpan = extension->SpanOfController;
    for (ioAddr = ioBase; ioAddr < ioBase + ioSpan; ++ioAddr) {
        mask = 1 << (ioAddr & 7);
        offset = ioAddr >> 3;
        if (offset >= IOPM_SIZE) break;
        *(*IOPM_local + offset) &= ~mask; // clear IOPM bit (allow access)
    }
    Ke386SetIoAccessMap(1, IOPM_local);

    // have we been there, done that already?
    if (extension->OpenCount > 0) {
        pIrp->IoStatus.Status = STATUS_SUCCESS;
        IoCompleteRequest(pIrp, IO_NO_INCREMENT);
        ++extension->OpenCount;  // bump usage count
        return STATUS_SUCCESS;
    }

    // initialize kernel event and build new device request
    KeInitializeEvent(&event, NotificationEvent, FALSE);
    irp = IoBuildDeviceIoControlRequest
                 (IOCTL_INTERNAL_PARALLEL_PORT_ALLOCATE,
                 extension->PortDeviceObject, NULL, 0, NULL, 0, TRUE, &event,
                 &ioStatus);
    if (!irp) {
        return STATUS_INSUFFICIENT_RESOURCES;
    }

    // call port driver
    Status = IoCallDriver(extension->PortDeviceObject, irp);
    if (!NT_SUCCESS(Status)) {
        pIrp->IoStatus.Status = STATUS_UNSUCCESSFUL;
        IoCancelIrp(irp);
        return Status;
    }

    // wait a bit to acquire port
    liTimeout.QuadPart = -(20 * 1000 * 1000);  // units of 100 nsec
    Status = KeWaitForSingleObject(
                       &event, Executive, KernelMode, FALSE, &liTimeout);
    if (Status == STATUS_TIMEOUT) {
        IoCancelIrp(irp);
        pIrp->IoStatus.Status = STATUS_ACCESS_DENIED;
        IoCompleteRequest(pIrp, IO_NO_INCREMENT);
        return STATUS_ACCESS_DENIED;
    } else {
        pIrp->IoStatus.Status = STATUS_SUCCESS;
        IoCompleteRequest(pIrp, IO_NO_INCREMENT);
        ++extension->OpenCount;  // bump usage count
        return STATUS_SUCCESS;
    }
}

//
//  Driver close (i.e., CloseHandle) handler
//
NTSTATUS PrivIoXPClose(IN PDEVICE_OBJECT DeviceObject, IN PIRP pIrp)
{
    PDEVICE_EXTENSION extension = DeviceObject->DeviceExtension;

    --extension->OpenCount;  // decrement usage count
    if (extension->OpenCount < 0) extension->OpenCount = 0;

    pIrp->IoStatus.Information = 0;
    pIrp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest(pIrp, IO_NO_INCREMENT);

    // do nothing until the last user closes
    if (extension->OpenCount == 0) {
        extension->FreePort(extension->FreePortContext);  // free port device
    }
    return (STATUS_SUCCESS);
}

//
//  Driver I/O Read Request handler
//
NTSTATUS PrivIoXPRead(IN PDEVICE_OBJECT DeviceObject, IN PIRP pIrp)
{
    PDEVICE_EXTENSION extension = DeviceObject->DeviceExtension;
    PIO_STACK_LOCATION pIrpStack;
    ULONG *pdwIoBuffer;
    ULONG nReadLen;
    NTSTATUS Status;

    pIrpStack = IoGetCurrentIrpStackLocation(pIrp);
    nReadLen = pIrpStack->Parameters.Read.Length;
    pdwIoBuffer = (pIrp->AssociatedIrp.SystemBuffer);
    pIrp->IoStatus.Information = 0;  // default to return no data

    if (nReadLen - pIrp->IoStatus.Information > sizeof(ULONG)) {
        // return base I/O address
        *pdwIoBuffer++ = (ULONG)(extension->Controller);
        pIrp->IoStatus.Information += sizeof(ULONG);
        if (nReadLen - pIrp->IoStatus.Information > sizeof(ULONG)) {
            // return I/O span
            *pdwIoBuffer++ = (ULONG)(extension->SpanOfController);
            pIrp->IoStatus.Information += sizeof(ULONG);
            if (nReadLen - pIrp->IoStatus.Information > sizeof(ULONG)) {
                // return driver open count
                *pdwIoBuffer++ = (ULONG)(extension->OpenCount);
                pIrp->IoStatus.Information += sizeof(ULONG);
            }
        }
    }
    pIrp->IoStatus.Status = STATUS_SUCCESS;
    IoCompleteRequest(pIrp, IO_NO_INCREMENT);
    return STATUS_SUCCESS;
}

//
//  Driver I/O control (i.e., DeviceIOCtl) handler
//
NTSTATUS PrivIoXPIOCtl(IN PDEVICE_OBJECT DeviceObject, IN PIRP pIrp)
{
    PDEVICE_EXTENSION extension = DeviceObject->DeviceExtension;
    PIO_STACK_LOCATION pIrpStack;
    KIRQL kCancelSpin;
    BOOLEAN bAllocPort;
    ULONG ulDataBufferLen;

    PVOID pIoBuffer;
    PUCHAR pucIoBuffer;
    PUSHORT pusIoBuffer;
    PULONG pulIoBuffer;

    ULONG ulInBufferLen;
    ULONG ulOutBufferLen;
    ULONG offset;
    UCHAR value;
    UCHAR mask;
    ULONG ProcessID;
    struct _EPROCESS *Eprocess;
    NTSTATUS Status;

    pIrpStack = IoGetCurrentIrpStackLocation(pIrp);
    ulInBufferLen = pIrpStack->Parameters.DeviceIoControl.InputBufferLength;
    ulOutBufferLen = pIrpStack->Parameters.DeviceIoControl.OutputBufferLength;
    pIoBuffer = pIrp->AssociatedIrp.SystemBuffer;

    pucIoBuffer = (PUCHAR)pIoBuffer;  // alias pIoBuffer
    pusIoBuffer = (PUSHORT)pIoBuffer;
    pulIoBuffer = (PULONG)pIoBuffer;

    pIrp->IoStatus.Information = 0;  // default to return no data

    switch (pIrpStack->Parameters.DeviceIoControl.IoControlCode) {

    case IOCTL_PRIVIOXP_WRITE_PORT_DATA:

        // write data
        ulDataBufferLen = sizeof(UCHAR);
        if (ulInBufferLen < ulDataBufferLen) {
            Status = STATUS_INVALID_PARAMETER;
            break;
        }
        WRITE_PORT_UCHAR((PUCHAR)(
                    (ULONG_PTR) extension->Controller + PPORT_DATA_OFFSET),
                    *(PUCHAR) pIoBuffer);
        Status = STATUS_SUCCESS;
        break;

    case IOCTL_PRIVIOXP_READ_PORT_STATUS:

        // read the data on the status registers
        ulDataBufferLen = sizeof(UCHAR);
        if (ulOutBufferLen < ulDataBufferLen) {
            Status = STATUS_INVALID_PARAMETER;
            break;
        }
        *(PUCHAR)pIoBuffer = READ_PORT_UCHAR(
                                (PUCHAR) ((ULONG_PTR) extension->Controller +
                                                     PPORT_STATUS_OFFSET));
        pIrp->IoStatus.Information = ulDataBufferLen;
        Status = STATUS_SUCCESS;
        break;

    case IOCTL_PRIVIOXP_READ_PORT_BASE:

        ulDataBufferLen = sizeof(ULONG);
        if (ulOutBufferLen < ulDataBufferLen) {
            Status = STATUS_INVALID_PARAMETER;
            break;
        }
        // return base I/O address of parallel port
        *((PULONG)pIoBuffer) = (ULONG)(extension->Controller);
        pIrp->IoStatus.Information = ulDataBufferLen;
        Status = STATUS_SUCCESS;
        break;

    case IOCTL_PRIVIOXP_READ_OPEN_COUNT:

        ulDataBufferLen = sizeof(LONG);
        if (ulOutBufferLen < ulDataBufferLen) {
            Status = STATUS_INVALID_PARAMETER;
            break;
        }
        // return "open count" (we count as one!)
        *((PLONG)pIoBuffer) = (LONG)(extension->OpenCount);
        pIrp->IoStatus.Information = ulDataBufferLen;
        Status = STATUS_SUCCESS;
        break;

    case IOCTL_IOPM_RESTRICT_ALL_ACCESS:

        // start with current I/O permissions from caller
        Ke386IoSetAccessProcess(PsGetCurrentProcess(), 1);
        Ke386QueryIoAccessMap(1, IOPM_local);
        RtlFillMemory(IOPM_local, sizeof(IOPM), 0xff);  // no more access
        Ke386SetIoAccessMap(1, IOPM_local);  // write new I/O permissions
        Status = STATUS_SUCCESS;
        break;

    case IOCTL_IOPM_ALLOW_ALL_ACCESS:

        // start with current I/O permissions from caller
        Ke386IoSetAccessProcess(PsGetCurrentProcess(), 1);
        Ke386QueryIoAccessMap(1, IOPM_local);
        RtlZeroMemory(IOPM_local, sizeof(IOPM));  // allow unrestricted access
        Ke386SetIoAccessMap(1, IOPM_local);  // write new I/O permissions
        Status = STATUS_SUCCESS;
        break;

    case IOCTL_SET_IOPM:

        if (pIrpStack->Parameters.DeviceIoControl.InputBufferLength < 4) {
            Status = STATUS_BUFFER_TOO_SMALL;
            break;
        }
        mask = 1 << (pulIoBuffer[0] & 7);
        offset = pulIoBuffer[0] >> 3;
        if (offset >= IOPM_SIZE) {
            Status = STATUS_ARRAY_BOUNDS_EXCEEDED;
            break;
        }

        // start with current I/O permissions from caller
        Ke386IoSetAccessProcess(PsGetCurrentProcess(), 1);
        Ke386QueryIoAccessMap(1, IOPM_local);
        *(*IOPM_local + offset) &= ~mask;   // clear IOPM bit (allow access)
        Ke386SetIoAccessMap(1, IOPM_local); // write new I/O permissions
        Status = STATUS_SUCCESS;
        break;

    case IOCTL_CLEAR_IOPM:

        if (pIrpStack->Parameters.DeviceIoControl.InputBufferLength < 4) {
            Status = STATUS_BUFFER_TOO_SMALL;
            break;
        }
        mask = 1 << (pulIoBuffer[0] & 7);
        offset = pulIoBuffer[0] >> 3;
        if (offset >= IOPM_SIZE) {
            Status = STATUS_ARRAY_BOUNDS_EXCEEDED;
            break;
        }
        value = mask;                // forbid I/O

        // start with current I/O permissions from caller
        Ke386IoSetAccessProcess(PsGetCurrentProcess(), 1);
        Ke386QueryIoAccessMap(1, IOPM_local);
        *(*IOPM_local + offset) &= ~mask;
        *(*IOPM_local + offset) |= value;    // set IOPM bit (no access)
        Ke386SetIoAccessMap(1, IOPM_local);  // write new I/O permissions
        Status = STATUS_SUCCESS;
        break;

    case IOCTL_ENABLE_IOPM_ON_PROCESSID:

        if (pIrpStack->Parameters.DeviceIoControl.InputBufferLength < 4) {
            Status = STATUS_BUFFER_TOO_SMALL;
            break;
        } else {
            // start with same I/O permissions as us
            Ke386IoSetAccessProcess(PsGetCurrentProcess(), 1);
            Ke386QueryIoAccessMap(1, IOPM_local);
            ProcessID = pulIoBuffer[0];
            Status = PsLookupProcessByProcessId(ProcessID, &Eprocess);
            if (Status != STATUS_SUCCESS) {
                Status = STATUS_INVALID_PARAMETER;
                break;
            }
            Ke386IoSetAccessProcess(Eprocess, 1); // switch to process IOPM
            Ke386SetIoAccessMap(1, IOPM_local);   // write new I/O permissions
            Status = STATUS_SUCCESS;
        }
        break;

    default:
        Status = STATUS_INVALID_PARAMETER;

    }

    // complete the request
    IoCompleteRequest(pIrp, IO_NO_INCREMENT);

    return Status;
}

//
//  Cancel IRP for a given device (entered with spin-lock!)
//
VOID PrivIoXPCancel(IN PDEVICE_OBJECT DeviceObject, IN PIRP pIrp)
{
    RemoveEntryList(&pIrp->Tail.Overlay.ListEntry);
    IoReleaseCancelSpinLock(pIrp->CancelIrql);
    pIrp->IoStatus.Information = 0;
    pIrp->IoStatus.Status = STATUS_CANCELLED;
    IoCompleteRequest(pIrp, IO_NO_INCREMENT);
}

//
//  cleanup any IRPs in work queue
//
NTSTATUS PrivIoXPCleanup(IN PDEVICE_OBJECT DeviceObject, IN PIRP pIrp)
{
    PDEVICE_EXTENSION extension = DeviceObject->DeviceExtension;
    KIRQL kCancelSpin;
    PLIST_ENTRY pHead;
    PIRP irp;

    IoAcquireCancelSpinLock(&kCancelSpin);  // need spin-lock to get started
    while (!IsListEmpty(&extension->WorkQueue)) {  // loop until queue empty
        pHead = RemoveHeadList(&extension->WorkQueue);
        irp = CONTAINING_RECORD(pHead, IRP, Tail.Overlay.ListEntry);
        irp->Cancel = TRUE;
        irp->CancelRoutine = NULL;
        irp->CancelIrql = kCancelSpin;
        PrivIoXPCancel(DeviceObject, irp);
        IoAcquireCancelSpinLock(&kCancelSpin);
    }
    IoReleaseCancelSpinLock(kCancelSpin);

    // we are done...
    pIrp->IoStatus.Status = STATUS_SUCCESS;
    pIrp->IoStatus.Information = 0;
    IoCompleteRequest(pIrp, IO_NO_INCREMENT);
    return (STATUS_SUCCESS);
}

//
//  Driver unload handler (clean-up all allocated devices)
//
VOID PrivIoXPUnload(IN PDRIVER_OBJECT DriverObject)
{
    PDEVICE_OBJECT currentDevice;
    PDEVICE_EXTENSION extension;

    while (currentDevice = DriverObject->DeviceObject) {
        extension = currentDevice->DeviceExtension;
        if (extension->CreatedSymbolicLink) {
            IoDeleteSymbolicLink(&extension->SymbolicLinkName);
            ExFreePool(extension->SymbolicLinkName.Buffer);
        }
        IoDeleteDevice(currentDevice);
    }
}

